/* Copyright (c) 2007 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

/* $Header:$ */

/* wrapper function for connect() which can lower the default system
   timeout upon socket connections. */

#define _GNU_SOURCE 1 		/* to get the RTLD_NEXT */
#include <dlfcn.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#define CONNECT_TIMEOUT "CONNECT_TIMEOUT"

#define DOJDEBUG 1

static int (*real_connect)(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) = 0;

// some code ripped from http://www.developerweb.net/forum/showthread.php?p=13486
int connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
#if DOJDEBUG
  fprintf(stderr, "connect_wrapper start\n");
#endif

  /* get out timeout in seconds */
  int timeout=10;
  if(getenv(CONNECT_TIMEOUT))
    {
      const int t=strtol(getenv(CONNECT_TIMEOUT), NULL, 0);
      if(t>0)
	timeout=t;
    }

  /* resolve real connect function */
  if(!real_connect)
    {
#if DOJDEBUG
      fprintf(stderr, "get real_connect symbol\n");
#endif
      dlerror();		/* clear last errors */
      real_connect=dlsym(RTLD_NEXT, "connect");
      if(!real_connect)
	{
	  const char *err=dlerror();
	  fprintf(stderr, "connect_wrapper(): could not lookup connect function: %s\n", err);
	  exit(EXIT_FAILURE);
	}
    }

  /* set socket to non-blocking */
#if DOJDEBUG
  fprintf(stderr, "set socket non-block\n");
#endif
  const int arg = fcntl(sockfd, F_GETFL, NULL);
  if(arg < 0)
    {
#if DOJDEBUG
      fprintf(stderr, "error during fcntl(F_GETFL): %s\n", strerror(errno));
#endif
      return -1;
    }
  if(fcntl(sockfd, F_SETFL, arg|O_NONBLOCK) < 0)
    {
#if DOJDEBUG
      fprintf(stderr, "error during fcntl(F_SETFL, %08X): %s\n", arg|O_NONBLOCK, strerror(errno));
#endif
      return -1;
    }

  /* not connect. Because socket is non-blocking real_connect() will return immediately. */
#if DOJDEBUG
  fprintf(stderr, "connect with %is timeout\n", timeout);
#endif
  int res=real_connect(sockfd, serv_addr, addrlen);
  if (res < 0)
    {
      /* when socket is trying to connect we have EINPROGRESS */
      if (errno == EINPROGRESS)
	{
	  /* now setup a select() call to wait for a connection */
	  fd_set myset;
	  struct timeval tv;
	  tv.tv_sec = timeout;
	  tv.tv_usec = 0;
	  FD_ZERO(&myset);
	  FD_SET(sockfd, &myset);
	  if ((res=select(sockfd+1, NULL, &myset, NULL, &tv)) > 0)
	    {
	      /* check that socket is connected */
	      int lon = sizeof(int), valopt;
	      if(getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon) < 0)
		{
#if DOJDEBUG
		  fprintf(stderr, "error during getsockopt(): %s\n", strerror(errno));
#endif
		  res=-1;
		}
	      if (valopt)
		{
		  /* we could not connect */
#if DOJDEBUG
		  fprintf(stderr, "Error in connection() %d - %s\n", valopt, strerror(valopt));
#endif
		  res=-1;
		  errno=valopt;
		}
	      else
		res=0;		/* connection success */
	    }
	  else if(res == 0)
	    {
	      /* timeout while connecting */
#if DOJDEBUG
	      fprintf(stderr, "Timeout on connect()\n");
#endif
	      errno=ETIMEDOUT;
	      res=-1;
	    }
	}
      else
	{
#if DOJDEBUG
	  fprintf(stderr, "Error connecting %d - %s\n", errno, strerror(errno));
#endif
	}
    }

#if DOJDEBUG
  fprintf(stderr, "restore socket state\n");
#endif
  if(fcntl(sockfd, F_SETFL, arg) < 0)
    {
#if DOJDEBUG
      fprintf(stderr, "error during fcntl(F_SETFL, %08X): %s\n", arg, strerror(errno));
#endif
    }

  return res;
}
